# Copyright 2022 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import http.client as http_client
import logging
import testresources
import uuid
import webob.dec

from oslo_config import cfg

from keystoneauth1 import exceptions as ksa_exceptions

from keystonemiddleware import oauth2_token
from keystonemiddleware.tests.unit.auth_token import base
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import BASE_URI
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import ENDPOINT_NOT_FOUND_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import ERROR_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import FAKE_ADMIN_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import FAKE_ADMIN_TOKEN_ID
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware\
    import FakeApp
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import FakeOsloCache
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import request_timeout_response
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import TIMEOUT_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import v3FakeApp
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import VERSION_LIST_v3
from keystonemiddleware.tests.unit import client_fixtures
from keystonemiddleware.tests.unit import utils


def get_authorization_header(token):
    return {'Authorization': f'Bearer {token}'}


class FakeOauth2TokenV3App(v3FakeApp):

    @webob.dec.wsgify
    def __call__(self, req):
        resp = webob.Response()
        resp.body = FakeApp.SUCCESS
        return resp


class BaseOauth2TokenMiddlewareTest(base.BaseAuthTokenTestCase):

    def setUp(self, expected_env=None, auth_version=None, fake_app=None):
        cfg.CONF.clear()
        super(BaseOauth2TokenMiddlewareTest, self).setUp()

        self.logger = self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
        self.useFixture(fixtures.MockPatchObject(oauth2_token.OAuth2Protocol,
                                                 '_create_oslo_cache',
                                                 return_value=FakeOsloCache))
        self.expected_env = expected_env or dict()
        self.fake_app = fake_app or FakeApp
        self.middleware = None

        self.conf = {
            'identity_uri': 'https://keystone.example.com:1234/testadmin/',
            'auth_version': auth_version,
            'www_authenticate_uri': 'https://keystone.example.com:1234',
            'admin_user': uuid.uuid4().hex,
        }
        self.auth_version = auth_version

    def call_middleware(self, **kwargs):
        return self.call(self.middleware, **kwargs)

    def set_middleware(self, expected_env=None, conf=None):
        """Configure the class ready to call the oauth2_token middleware.

        Set up the various fake items needed to run the middleware.
        Individual tests that need to further refine these can call this
        function to override the class defaults.

        """
        if conf:
            self.conf.update(conf)

        if expected_env:
            self.expected_env.update(expected_env)
        self.middleware = oauth2_token.OAuth2Protocol(
            self.fake_app(self.expected_env), self.conf)

    def call(self, middleware, method='GET', path='/', headers=None,
             expected_status=http_client.OK,
             expected_body_string=None):
        req = webob.Request.blank(path)
        req.method = method

        for k, v in (headers or {}).items():
            req.headers[k] = v

        resp = req.get_response(middleware)
        self.assertEqual(expected_status, resp.status_int)
        if expected_body_string:
            self.assertIn(expected_body_string, str(resp.body))
        resp.request = req
        return resp


class Oauth2TokenMiddlewareTest(BaseOauth2TokenMiddlewareTest,
                                testresources.ResourcedTestCase):

    resources = [('examples', client_fixtures.EXAMPLES_RESOURCE)]

    def setUp(self):
        super(Oauth2TokenMiddlewareTest, self).setUp(
            auth_version='v3.0',
            fake_app=FakeOauth2TokenV3App)
        self.requests_mock.post('%s/v2.0/tokens' % BASE_URI,
                                text=FAKE_ADMIN_TOKEN)
        self.requests_mock.get(BASE_URI,
                               json=VERSION_LIST_v3,
                               status_code=300)
        self.requests_mock.get('%s/v3/auth/tokens' % BASE_URI,
                               text=self.token_response,
                               headers={'X-Subject-Token': uuid.uuid4().hex})
        self.set_middleware()

    def token_response(self, request, context):
        auth_id = request.headers.get('X-Auth-Token')
        token_id = request.headers.get('X-Subject-Token')
        self.assertEqual(auth_id, FAKE_ADMIN_TOKEN_ID)

        if token_id == ERROR_TOKEN:
            msg = "Network connection refused."
            raise ksa_exceptions.ConnectFailure(msg)
        if token_id == ENDPOINT_NOT_FOUND_TOKEN:
            raise ksa_exceptions.EndpointNotFound()
        if token_id == TIMEOUT_TOKEN:
            request_timeout_response(request, context)

        try:
            response = self.examples.JSON_TOKEN_RESPONSES[token_id]
        except KeyError:
            response = ""
            context.status_code = 404

        return response

    def test_app_cred_token_without_access_rules(self):
        self.set_middleware(conf={'service_type': 'compute'})
        token = self.examples.v3_APP_CRED_TOKEN
        token_data = self.examples.TOKEN_RESPONSES[token]
        resp = self.call_middleware(headers=get_authorization_header(token))
        self.assertEqual(FakeApp.SUCCESS, resp.body)
        token_auth = resp.request.environ['keystone.token_auth']
        self.assertEqual(token_data.application_credential_id,
                         token_auth.user.application_credential_id)

    def test_app_cred_access_rules_token(self):
        self.set_middleware(conf={'service_type': 'compute'})
        token = self.examples.v3_APP_CRED_ACCESS_RULES
        token_data = self.examples.TOKEN_RESPONSES[token]
        resp = self.call_middleware(headers=get_authorization_header(token),
                                    expected_status=200,
                                    method='GET', path='/v2.1/servers')
        token_auth = resp.request.environ['keystone.token_auth']
        self.assertEqual(token_data.application_credential_id,
                         token_auth.user.application_credential_id)
        self.assertEqual(token_data.application_credential_access_rules,
                         token_auth.user.application_credential_access_rules)
        resp = self.call_middleware(headers=get_authorization_header(token),
                                    expected_status=401,
                                    method='GET',
                                    path='/v2.1/servers/someuuid')
        self.assertEqual(token_data.application_credential_id,
                         token_auth.user.application_credential_id)
        self.assertEqual(token_data.application_credential_access_rules,
                         token_auth.user.application_credential_access_rules)

    def test_app_cred_no_access_rules_token(self):
        self.set_middleware(conf={'service_type': 'compute'})
        token = self.examples.v3_APP_CRED_EMPTY_ACCESS_RULES
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/v2.1/servers')

    def test_app_cred_matching_rules(self):
        self.set_middleware(conf={'service_type': 'compute'})
        token = self.examples.v3_APP_CRED_MATCHING_RULES
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=200,
                             method='GET', path='/v2.1/servers/foobar')
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/v2.1/servers/foobar/barfoo')
        self.set_middleware(conf={'service_type': 'image'})
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=200,
                             method='GET', path='/v2/images/foobar')
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/v2/images/foobar/barfoo')
        self.set_middleware(conf={'service_type': 'identity'})
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=200,
                             method='GET',
                             path='/v3/projects/123/users/456/roles/member')
        self.set_middleware(conf={'service_type': 'block-storage'})
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=200,
                             method='GET', path='/v3/123/types/456')
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/v3/123/types')
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/v2/123/types/456')
        self.set_middleware(conf={'service_type': 'object-store'})
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=200,
                             method='GET', path='/v1/1/2/3')
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/v1/1/2')
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/v2/1/2')
        self.call_middleware(headers=get_authorization_header(token),
                             expected_status=401,
                             method='GET', path='/info')

    def test_request_no_token(self):
        resp = self.call_middleware(expected_status=401)
        self.assertEqual('Keystone uri="https://keystone.example.com:1234"',
                         resp.headers['WWW-Authenticate'])

    def test_request_blank_token(self):
        resp = self.call_middleware(headers=get_authorization_header(''),
                                    expected_status=401)
        self.assertEqual('Keystone uri="https://keystone.example.com:1234"',
                         resp.headers['WWW-Authenticate'])

    def test_request_not_app_cred_token(self):
        self.call_middleware(
            headers=get_authorization_header(
                self.examples.v3_UUID_TOKEN_DEFAULT),
            expected_status=200)

    def _get_cached_token(self, token):
        return self.middleware._token_cache.get(token)

    def assert_valid_last_url(self, token_id):
        self.assertLastPath('/v3/auth/tokens')

    def assertLastPath(self, path):
        if path:
            self.assertEqual(BASE_URI + path,
                             self.requests_mock.last_request.url)
        else:
            self.assertIsNone(self.requests_mock.last_request)

    def test_http_error_not_cached_token(self):
        """Test to don't cache token as invalid on network errors.

        We use UUID tokens since they are the easiest one to reach
        get_http_connection.
        """
        self.set_middleware(conf={'http_request_max_retries': '0'})
        self.call_middleware(headers=get_authorization_header(ERROR_TOKEN),
                             expected_status=503)
        self.assertIsNone(self._get_cached_token(ERROR_TOKEN))
        self.assert_valid_last_url(ERROR_TOKEN)


class FilterFactoryTest(utils.BaseTestCase):

    def test_filter_factory(self):
        conf = {}
        auth_filter = oauth2_token.filter_factory(conf)
        m = auth_filter(FakeOauth2TokenV3App())
        self.assertIsInstance(m, oauth2_token.OAuth2Protocol)



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import datetime
import hashlib
import os
import ssl
import uuid

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography import x509
import fixtures
from keystoneauth1 import fixture
from oslo_serialization import jsonutils
import testresources


TESTDIR = os.path.dirname(os.path.abspath(__file__))
ROOTDIR = os.path.normpath(os.path.join(TESTDIR, '..', '..', '..'))


class Examples(fixtures.Fixture):
    """Example tokens and certs loaded from the examples directory.

    To use this class correctly, the module needs to override the test suite
    class to use testresources.OptimisingTestSuite (otherwise the files will
    be read on every test). This is done by defining a load_tests function
    in the module, like this:

    def load_tests(loader, tests, pattern):
        return testresources.OptimisingTestSuite(tests)

    (see http://docs.python.org/2/library/unittest.html#load-tests-protocol )

    """

    def setUp(self):
        super(Examples, self).setUp()

        self.KERBEROS_BIND = 'USER@REALM'
        self.SERVICE_KERBEROS_BIND = 'SERVICE_USER@SERVICE_REALM'

        self.UUID_TOKEN_DEFAULT = "ec6c0710ec2f471498484c1b53ab4f9d"
        self.UUID_TOKEN_NO_SERVICE_CATALOG = '8286720fbe4941e69fa8241723bb02df'
        self.UUID_TOKEN_UNSCOPED = '731f903721c14827be7b2dc912af7776'
        self.UUID_TOKEN_BIND = '3fc54048ad64405c98225ce0897af7c5'
        self.UUID_TOKEN_UNKNOWN_BIND = '8885fdf4d42e4fb9879e6379fa1eaf48'
        self.v3_UUID_TOKEN_DEFAULT = '5603457654b346fdbb93437bfe76f2f1'
        self.v3_UUID_TOKEN_UNSCOPED = 'd34835fdaec447e695a0a024d84f8d79'
        self.v3_UUID_TOKEN_DOMAIN_SCOPED = 'e8a7b63aaa4449f38f0c5c05c3581792'
        self.v3_UUID_TOKEN_BIND = '2f61f73e1c854cbb9534c487f9bd63c2'
        self.v3_UUID_TOKEN_UNKNOWN_BIND = '7ed9781b62cd4880b8d8c6788ab1d1e2'
        self.v3_SYSTEM_SCOPED_TOKEN = '9ca6e88364b6418a88ffc02e6a24afd8'

        self.UUID_SERVICE_TOKEN_DEFAULT = 'fe4c0710ec2f492748596c1b53ab124'
        self.UUID_SERVICE_TOKEN_BIND = '5e43439613d34a13a7e03b2762bd08ab'
        self.v3_UUID_SERVICE_TOKEN_DEFAULT = 'g431071bbc2f492748596c1b53cb229'
        self.v3_UUID_SERVICE_TOKEN_BIND = 'be705e4426d0449a89e35ae21c380a05'
        self.v3_NOT_IS_ADMIN_PROJECT = uuid.uuid4().hex

        self.v3_APP_CRED_TOKEN = '6f506fa9641448bbaecbd12dd30678a9'
        self.v3_APP_CRED_ACCESS_RULES = 'c417747898c44629b08791f2579e40a5'
        self.v3_APP_CRED_EMPTY_ACCESS_RULES = 'c75905c307f04fdd9979126582d7aae'
        self.v3_APP_CRED_MATCHING_RULES = 'ad49decc7106489d95ca9ed874b6cb66'

        self.v3_OAUTH2_CREDENTIAL = uuid.uuid4().hex
        self.V3_OAUTH2_MTLS_CERTIFICATE = self._create_pem_certificate(
            self._create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organization_name='fujitsu',
                organizational_unit_name='test',
                common_name='root'
            )
        )
        self.V3_OAUTH2_MTLS_CERTIFICATE_DIFF = self._create_pem_certificate(
            self._create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organization_name='fujitsu',
                organizational_unit_name='test',
                common_name='diff'
            )
        )

        # JSON responses keyed by token ID
        self.TOKEN_RESPONSES = {}

        # basic values
        PROJECT_ID = 'tenant_id1'
        PROJECT_NAME = 'tenant_name1'
        USER_ID = 'user_id1'
        USER_NAME = 'user_name1'
        DOMAIN_ID = 'domain_id1'
        DOMAIN_NAME = 'domain_name1'
        ROLE_NAME1 = 'role1'
        ROLE_NAME2 = 'role2'

        SERVICE_PROJECT_ID = 'service_project_id1'
        SERVICE_PROJECT_NAME = 'service_project_name1'
        SERVICE_USER_ID = 'service_user_id1'
        SERVICE_USER_NAME = 'service_user_name1'
        SERVICE_DOMAIN_ID = 'service_domain_id1'
        SERVICE_DOMAIN_NAME = 'service_domain_name1'
        SERVICE_ROLE_NAME1 = 'service'
        SERVICE_ROLE_NAME2 = 'service_role2'

        APP_CRED_ID = 'app_cred_id1'

        self.SERVICE_TYPE = 'identity'
        self.UNVERSIONED_SERVICE_URL = 'https://keystone.example.com:1234/'
        self.SERVICE_URL = self.UNVERSIONED_SERVICE_URL + 'v2.0'

        # Generated V2 Tokens

        token = fixture.V2Token(token_id=self.UUID_TOKEN_DEFAULT,
                                tenant_id=PROJECT_ID,
                                tenant_name=PROJECT_NAME,
                                user_id=USER_ID,
                                user_name=USER_NAME)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint(public=self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.UUID_TOKEN_DEFAULT] = token

        token = fixture.V2Token(token_id=self.UUID_TOKEN_UNSCOPED,
                                user_id=USER_ID,
                                user_name=USER_NAME)
        self.TOKEN_RESPONSES[self.UUID_TOKEN_UNSCOPED] = token

        token = fixture.V2Token(token_id='valid-token',
                                tenant_id=PROJECT_ID,
                                tenant_name=PROJECT_NAME,
                                user_id=USER_ID,
                                user_name=USER_NAME)
        token.add_role(ROLE_NAME1)
        token.add_role(ROLE_NAME2)
        self.TOKEN_RESPONSES[self.UUID_TOKEN_NO_SERVICE_CATALOG] = token

        token = fixture.V2Token(token_id=self.UUID_TOKEN_BIND,
                                tenant_id=PROJECT_ID,
                                tenant_name=PROJECT_NAME,
                                user_id=USER_ID,
                                user_name=USER_NAME)
        token.add_role(ROLE_NAME1)
        token.add_role(ROLE_NAME2)
        token['access']['token']['bind'] = {'kerberos': self.KERBEROS_BIND}
        self.TOKEN_RESPONSES[self.UUID_TOKEN_BIND] = token

        token = fixture.V2Token(token_id=self.UUID_SERVICE_TOKEN_BIND,
                                tenant_id=SERVICE_PROJECT_ID,
                                tenant_name=SERVICE_PROJECT_NAME,
                                user_id=SERVICE_USER_ID,
                                user_name=SERVICE_USER_NAME)
        token.add_role(SERVICE_ROLE_NAME1)
        token.add_role(SERVICE_ROLE_NAME2)
        token['access']['token']['bind'] = {
            'kerberos': self.SERVICE_KERBEROS_BIND}
        self.TOKEN_RESPONSES[self.UUID_SERVICE_TOKEN_BIND] = token

        token = fixture.V2Token(token_id=self.UUID_TOKEN_UNKNOWN_BIND,
                                tenant_id=PROJECT_ID,
                                tenant_name=PROJECT_NAME,
                                user_id=USER_ID,
                                user_name=USER_NAME)
        token.add_role(ROLE_NAME1)
        token.add_role(ROLE_NAME2)
        token['access']['token']['bind'] = {'FOO': 'BAR'}
        self.TOKEN_RESPONSES[self.UUID_TOKEN_UNKNOWN_BIND] = token

        token = fixture.V2Token(token_id=self.UUID_SERVICE_TOKEN_DEFAULT,
                                tenant_id=SERVICE_PROJECT_ID,
                                tenant_name=SERVICE_PROJECT_NAME,
                                user_id=SERVICE_USER_ID,
                                user_name=SERVICE_USER_NAME)
        token.add_role(name=SERVICE_ROLE_NAME1)
        token.add_role(name=SERVICE_ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint(public=self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.UUID_SERVICE_TOKEN_DEFAULT] = token

        # Generated V3 Tokens

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME,
                                project_id=PROJECT_ID,
                                project_name=PROJECT_NAME,
                                project_domain_id=DOMAIN_ID,
                                project_domain_name=DOMAIN_NAME)
        token.add_role(id=ROLE_NAME1, name=ROLE_NAME1)
        token.add_role(id=ROLE_NAME2, name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.v3_UUID_TOKEN_DEFAULT] = token

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME)
        self.TOKEN_RESPONSES[self.v3_UUID_TOKEN_UNSCOPED] = token

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME)
        token.system = {'all': True}
        token.add_role(id=ROLE_NAME1, name=ROLE_NAME1)
        token.add_role(id=ROLE_NAME2, name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.v3_SYSTEM_SCOPED_TOKEN] = token

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME,
                                domain_id=DOMAIN_ID,
                                domain_name=DOMAIN_NAME)
        token.add_role(id=ROLE_NAME1, name=ROLE_NAME1)
        token.add_role(id=ROLE_NAME2, name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.v3_UUID_TOKEN_DOMAIN_SCOPED] = token

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME,
                                project_id=PROJECT_ID,
                                project_name=PROJECT_NAME,
                                project_domain_id=DOMAIN_ID,
                                project_domain_name=DOMAIN_NAME)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME,
                                project_id=PROJECT_ID,
                                project_name=PROJECT_NAME,
                                project_domain_id=DOMAIN_ID,
                                project_domain_name=DOMAIN_NAME)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        token['token']['bind'] = {'kerberos': self.KERBEROS_BIND}
        self.TOKEN_RESPONSES[self.v3_UUID_TOKEN_BIND] = token

        token = fixture.V3Token(user_id=SERVICE_USER_ID,
                                user_name=SERVICE_USER_NAME,
                                user_domain_id=SERVICE_DOMAIN_ID,
                                user_domain_name=SERVICE_DOMAIN_NAME,
                                project_id=SERVICE_PROJECT_ID,
                                project_name=SERVICE_PROJECT_NAME,
                                project_domain_id=SERVICE_DOMAIN_ID,
                                project_domain_name=SERVICE_DOMAIN_NAME)
        token.add_role(name=SERVICE_ROLE_NAME1)
        token.add_role(name=SERVICE_ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        token['token']['bind'] = {'kerberos': self.SERVICE_KERBEROS_BIND}
        self.TOKEN_RESPONSES[self.v3_UUID_SERVICE_TOKEN_BIND] = token

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME,
                                project_id=PROJECT_ID,
                                project_name=PROJECT_NAME,
                                project_domain_id=DOMAIN_ID,
                                project_domain_name=DOMAIN_NAME)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        token['token']['bind'] = {'FOO': 'BAR'}
        self.TOKEN_RESPONSES[self.v3_UUID_TOKEN_UNKNOWN_BIND] = token

        token = fixture.V3Token(user_id=SERVICE_USER_ID,
                                user_name=SERVICE_USER_NAME,
                                user_domain_id=SERVICE_DOMAIN_ID,
                                user_domain_name=SERVICE_DOMAIN_NAME,
                                project_id=SERVICE_PROJECT_ID,
                                project_name=SERVICE_PROJECT_NAME,
                                project_domain_id=SERVICE_DOMAIN_ID,
                                project_domain_name=SERVICE_DOMAIN_NAME)
        token.add_role(id=SERVICE_ROLE_NAME1,
                       name=SERVICE_ROLE_NAME1)
        token.add_role(id=SERVICE_ROLE_NAME2,
                       name=SERVICE_ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.v3_UUID_SERVICE_TOKEN_DEFAULT] = token

        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME,
                                project_id=PROJECT_ID,
                                project_name=PROJECT_NAME,
                                project_domain_id=DOMAIN_ID,
                                project_domain_name=DOMAIN_NAME,
                                is_admin_project=False)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.v3_NOT_IS_ADMIN_PROJECT] = token

        # Application credential token
        token = fixture.V3Token(user_id=USER_ID,
                                user_name=USER_NAME,
                                user_domain_id=DOMAIN_ID,
                                user_domain_name=DOMAIN_NAME,
                                project_id=PROJECT_ID,
                                project_name=PROJECT_NAME,
                                project_domain_id=DOMAIN_ID,
                                project_domain_name=DOMAIN_NAME,
                                application_credential_id=APP_CRED_ID)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        svc = token.add_service('compute')
        svc.add_endpoint('public', 'https://nova.openstack.example.org/v2.1')
        self.TOKEN_RESPONSES[self.v3_APP_CRED_TOKEN] = token

        # Application credential with access_rules token
        access_rules = [{
            'path': '/v2.1/servers',
            'method': 'GET',
            'service': 'compute'
        }]
        token = fixture.V3Token(
            user_id=USER_ID,
            user_name=USER_NAME,
            user_domain_id=DOMAIN_ID,
            user_domain_name=DOMAIN_NAME,
            project_id=PROJECT_ID,
            project_name=PROJECT_NAME,
            project_domain_id=DOMAIN_ID,
            project_domain_name=DOMAIN_NAME,
            application_credential_id=APP_CRED_ID,
            application_credential_access_rules=access_rules)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        svc = token.add_service('compute')
        svc.add_endpoint('public', 'https://nova.openstack.example.org')
        svc = token.add_service('image')
        svc.add_endpoint('public', 'https://glance.openstack.example.org')
        self.TOKEN_RESPONSES[self.v3_APP_CRED_ACCESS_RULES] = token

        # Application credential with explicitly empty access_rules
        access_rules = []
        token = fixture.V3Token(
            user_id=USER_ID,
            user_name=USER_NAME,
            user_domain_id=DOMAIN_ID,
            user_domain_name=DOMAIN_NAME,
            project_id=PROJECT_ID,
            project_name=PROJECT_NAME,
            project_domain_id=DOMAIN_ID,
            project_domain_name=DOMAIN_NAME,
            application_credential_id=APP_CRED_ID,
            application_credential_access_rules=access_rules)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        self.TOKEN_RESPONSES[self.v3_APP_CRED_EMPTY_ACCESS_RULES] = token

        # Application credential with matching rules
        access_rules = [
            {
                'path': '/v2.1/servers/{server_id}',
                'method': 'GET',
                'service': 'compute'
            },
            {
                'path': '/v2/images/*',
                'method': 'GET',
                'service': 'image'
            },
            {
                'path': '**',
                'method': 'GET',
                'service': 'identity'
            },
            {
                'path': '/v3/{project_id}/types/{volume_type_id}',
                'method': 'GET',
                'service': 'block-storage'
            },
            {
                'path': '/v1/*/*/*',
                'method': 'GET',
                'service': 'object-store'
            }
        ]
        token = fixture.V3Token(
            user_id=USER_ID,
            user_name=USER_NAME,
            user_domain_id=DOMAIN_ID,
            user_domain_name=DOMAIN_NAME,
            project_id=PROJECT_ID,
            project_name=PROJECT_NAME,
            project_domain_id=DOMAIN_ID,
            project_domain_name=DOMAIN_NAME,
            application_credential_id=APP_CRED_ID,
            application_credential_access_rules=access_rules)
        token.add_role(name=ROLE_NAME1)
        token.add_role(name=ROLE_NAME2)
        svc = token.add_service(self.SERVICE_TYPE)
        svc.add_endpoint('public', self.SERVICE_URL)
        svc = token.add_service('compute')
        svc.add_endpoint('public', 'https://nova.openstack.example.org')
        svc = token.add_service('image')
        svc.add_endpoint('public', 'https://glance.openstack.example.org')
        svc = token.add_service('block-storage')
        svc.add_endpoint('public', 'https://cinder.openstack.example.org')
        svc = token.add_service('object-store')
        svc.add_endpoint('public', 'https://swift.openstack.example.org')
        self.TOKEN_RESPONSES[self.v3_APP_CRED_MATCHING_RULES] = token

        # oauth2 credential token
        cert_pem = ssl.DER_cert_to_PEM_cert(self.V3_OAUTH2_MTLS_CERTIFICATE)
        thumb_sha256 = hashlib.sha256(cert_pem.encode('ascii')).digest()
        cert_thumb = base64.urlsafe_b64encode(thumb_sha256).decode('ascii')

        token = fixture.V3Token(
            methods=['oauth2_credential'],
            user_id=USER_ID,
            user_name=USER_NAME,
            project_id=PROJECT_ID,
            oauth2_thumbprint=cert_thumb,
        )
        self.TOKEN_RESPONSES[self.v3_OAUTH2_CREDENTIAL] = token

        self.JSON_TOKEN_RESPONSES = dict([(k, jsonutils.dumps(v)) for k, v in
                                          self.TOKEN_RESPONSES.items()])

    def _create_dn(
        self,
        common_name=None,
        locality_name=None,
        state_or_province_name=None,
        organization_name=None,
        organizational_unit_name=None,
        country_name=None,
        street_address=None,
        domain_component=None,
        user_id=None,
        email_address=None,
    ):
        oid = x509.NameOID
        attr = x509.NameAttribute
        dn = []
        if common_name:
            dn.append(attr(oid.COMMON_NAME, common_name))
        if locality_name:
            dn.append(attr(oid.LOCALITY_NAME, locality_name))
        if state_or_province_name:
            dn.append(attr(oid.STATE_OR_PROVINCE_NAME, state_or_province_name))
        if organization_name:
            dn.append(attr(oid.ORGANIZATION_NAME, organization_name))
        if organizational_unit_name:
            dn.append(
                attr(
                    oid.ORGANIZATIONAL_UNIT_NAME,
                    organizational_unit_name))
        if country_name:
            dn.append(attr(oid.COUNTRY_NAME, country_name))
        if street_address:
            dn.append(attr(oid.STREET_ADDRESS, street_address))
        if domain_component:
            dn.append(attr(oid.DOMAIN_COMPONENT, domain_component))
        if user_id:
            dn.append(attr(oid.USER_ID, user_id))
        if email_address:
            dn.append(attr(oid.EMAIL_ADDRESS, email_address))
        return x509.Name(dn)

    def _create_certificate(self, subject_dn, ca=None, ca_key=None):
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )
        issuer = ca.subject if ca else subject_dn
        if not ca_key:
            ca_key = private_key
        today = datetime.datetime.today()
        cert = x509.CertificateBuilder(
            issuer_name=issuer,
            subject_name=subject_dn,
            public_key=private_key.public_key(),
            serial_number=x509.random_serial_number(),
            not_valid_before=today,
            not_valid_after=today + datetime.timedelta(365, 0, 0),
        ).sign(ca_key, hashes.SHA256())

        return cert, private_key

    def _create_pem_certificate(self, subject_dn, ca=None, ca_key=None):
        cert, _ = self._create_certificate(subject_dn, ca=ca, ca_key=ca_key)
        return cert.public_bytes(Encoding.PEM)


EXAMPLES_RESOURCE = testresources.FixtureResource(Examples())



# Copyright 2022 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import base64
import hashlib
import ssl
import webob

from oslo_log import log as logging
from oslo_serialization import jsonutils

from keystonemiddleware.auth_token import _user_plugin
from keystonemiddleware.auth_token import AuthProtocol
from keystonemiddleware import exceptions
from keystonemiddleware.i18n import _


class OAuth2mTlsProtocol(AuthProtocol):
    """Middleware that handles OAuth2.0 mutual-TLS client authentication."""

    def __init__(self, app, conf):
        log = logging.getLogger(conf.get('log_name', __name__))
        log.info('Starting Keystone oauth2_mls_token middleware')
        super(OAuth2mTlsProtocol, self).__init__(app, conf)

    def _confirm_certificate_thumbprint(self, token_thumb, peer_cert):
        """Check if the thumbprint in the token is valid.

        :rtype: if the thumbprint is valid
        """
        try:
            cert_pem = ssl.DER_cert_to_PEM_cert(peer_cert)
            thumb_sha256 = hashlib.sha256(cert_pem.encode('ascii')).digest()
            cert_thumb = base64.urlsafe_b64encode(thumb_sha256).decode('ascii')
            if cert_thumb == token_thumb:
                is_valid = True
            else:
                self.log.info('The two thumbprints do not match.')
                is_valid = False
        except Exception as error:
            self.log.exception(error)
            is_valid = False
        return is_valid

    def _is_valid_access_token(self, request):
        """Check the OAuth2.0 certificate-bound access token.

        :param request: Incoming request
        :rtype: if the access token is valid
        """
        try:
            wsgi_input = request.environ.get("wsgi.input")
            if not wsgi_input:
                self.log.warn('Unable to obtain the client certificate.')
                return False
            sock = wsgi_input.get_socket()
            if not sock:
                self.log.warn('Unable to obtain the client certificate.')
                return False
            peer_cert = sock.getpeercert(binary_form=True)
            if not peer_cert:
                self.log.warn('Unable to obtain the client certificate.')
                return False
        except Exception as error:
            self.log.warn('Unable to obtain the client certificate. %s' %
                          str(error))
            return False

        access_token = None
        if (request.authorization and
                request.authorization.authtype == 'Bearer'):
            access_token = request.authorization.params

        if not access_token:
            self.log.info('Unable to obtain the token.')
            return False

        try:
            token_data, user_auth_ref = self._do_fetch_token(
                access_token, allow_expired=False)
            self._validate_token(user_auth_ref, allow_expired=False)
            token = token_data.get('token')
            oauth2_cred = token.get('oauth2_credential')
            if not oauth2_cred:
                self.log.info(
                    'Invalid OAuth2.0 certificate-bound access token: '
                    'The token is not an OAuth2.0 credential access token.')
                return False

            token_thumb = oauth2_cred.get("x5t#S256")
            if self._confirm_certificate_thumbprint(token_thumb, peer_cert):
                self._confirm_token_bind(user_auth_ref, request)
                request.token_info = token_data
                request.token_auth = _user_plugin.UserAuthPlugin(
                    user_auth_ref, None)
                return True
            else:
                self.log.info(
                    'Invalid OAuth2.0 certificate-bound access token: '
                    'the access token dose not match the client certificate.')
                return False
        except exceptions.KeystoneMiddlewareException as err:
            self.log.info('Invalid OAuth2.0 certificate-bound access token: %s'
                          % str(err))
            return False

    def process_request(self, request):
        """Process request.

        :param request: Incoming request
        :type request: _request.AuthTokenRequest
        """
        request.remove_auth_headers()
        self._token_cache.initialize(request.environ)
        if (not self._is_valid_access_token(request)
                or "keystone.token_info" not in request.environ
                or "token" not in request.environ["keystone.token_info"]):
            self.log.info('Rejecting request')
            message = _('The request you have made requires authentication.')
            body = {'error': {
                'code': 401,
                'title': 'Unauthorized',
                'message': message,
            }}
            raise webob.exc.HTTPUnauthorized(
                body=jsonutils.dumps(body),
                headers=self._reject_auth_headers,
                charset='UTF-8',
                content_type='application/json')

        request.set_user_headers(request.token_auth.user)
        request.set_service_catalog_headers(request.token_auth.user)
        request.token_auth._auth = self._auth
        request.token_auth._session = self._session
        self.log.debug('Accepting request and inited all env fields.')


def filter_factory(global_conf, **local_conf):
    """Return a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_filter(app):
        return OAuth2mTlsProtocol(app, conf)

    return auth_filter



# Copyright 2022 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import webob

from oslo_log import log as logging
from oslo_serialization import jsonutils

from keystonemiddleware.auth_token import _user_plugin
from keystonemiddleware.auth_token import AuthProtocol
from keystonemiddleware import exceptions
from keystonemiddleware.i18n import _


_LOG = logging.getLogger(__name__)


class OAuth2Protocol(AuthProtocol):
    """Middleware that handles OAuth2.0 client credentials authentication."""

    def __init__(self, app, conf):
        log = logging.getLogger(conf.get('log_name', __name__))
        log.info('Starting Keystone oauth2_token middleware')
        super(OAuth2Protocol, self).__init__(app, conf)

    def _is_valid_access_token(self, request):
        """Check if the request contains an OAuth2.0 access token.

        :param request: Incoming request
        :type request: _request.AuthTokenRequest
        """
        access_token = None
        if (request.authorization and
                request.authorization.authtype == 'Bearer'):
            access_token = request.authorization.params

        if access_token:
            try:
                token_data, user_auth_ref = self._do_fetch_token(
                    access_token, allow_expired=False)
                self._validate_token(user_auth_ref,
                                     allow_expired=False)
                token = token_data['token']
                self.validate_allowed_request(request, token)
                self._confirm_token_bind(user_auth_ref, request)
                request.token_info = token_data
                request.token_auth = _user_plugin.UserAuthPlugin(
                    user_auth_ref, None)
                return True
            except exceptions.KeystoneMiddlewareException as err:
                _LOG.info('Invalid OAuth2.0 access token: %s' % str(err))
        return False

    def process_request(self, request):
        """Process request.

        :param request: Incoming request
        :type request: _request.AuthTokenRequest
        """
        request.remove_auth_headers()
        self._token_cache.initialize(request.environ)
        if (not self._is_valid_access_token(request)
                or "keystone.token_info" not in request.environ
                or "token" not in request.environ["keystone.token_info"]):
            _LOG.info('Rejecting request')
            message = _('The request you have made requires authentication.')
            body = {'error': {
                'code': 401,
                'title': 'Unauthorized',
                'message': message,
            }}
            raise webob.exc.HTTPUnauthorized(
                body=jsonutils.dumps(body),
                headers=self._reject_auth_headers,
                charset='UTF-8',
                content_type='application/json')

        request.set_user_headers(request.token_auth.user)
        request.set_service_catalog_headers(request.token_auth.user)
        request.token_auth._auth = self._auth
        request.token_auth._session = self._session


def filter_factory(global_conf, **local_conf):
    """Return a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_filter(app):
        return OAuth2Protocol(app, conf)

    return auth_filter



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urllib.parse

from keystoneauth1 import discover
from keystoneauth1 import exceptions as ksa_exceptions
from keystoneauth1 import plugin
from keystoneclient.v3 import client as v3_client

from keystonemiddleware.auth_token import _auth
from keystonemiddleware.auth_token import _exceptions as ksm_exceptions
from keystonemiddleware.i18n import _

ACCESS_RULES_SUPPORT = '1'


class _RequestStrategy(object):

    AUTH_VERSION = None

    def __init__(self, adap, include_service_catalog=None,
                 requested_auth_interface=None):
        self._include_service_catalog = include_service_catalog
        self._requested_auth_interface = requested_auth_interface

    def verify_token(self, user_token, allow_expired=False):
        pass


class _V3RequestStrategy(_RequestStrategy):

    AUTH_VERSION = (3, 0)

    def __init__(self, adap, **kwargs):
        super(_V3RequestStrategy, self).__init__(adap, **kwargs)
        client_args = {'session': adap}
        if self._requested_auth_interface:
            client_args['interface'] = self._requested_auth_interface
        self._client = v3_client.Client(**client_args)

    def verify_token(self, token, allow_expired=False):
        auth_ref = self._client.tokens.validate(
            token,
            include_catalog=self._include_service_catalog,
            allow_expired=allow_expired,
            access_rules_support=ACCESS_RULES_SUPPORT)

        if not auth_ref:
            msg = _('Failed to fetch token data from identity server')
            raise ksm_exceptions.InvalidToken(msg)

        return {'token': auth_ref}


_REQUEST_STRATEGIES = [_V3RequestStrategy]


class IdentityServer(object):
    """Base class for operations on the Identity API server.

    The auth_token middleware needs to communicate with the Identity API server
    to validate tokens. This class encapsulates the data and methods to perform
    the operations.

    """

    def __init__(self, log, adap, include_service_catalog=None,
                 requested_auth_version=None, requested_auth_interface=None):
        self._LOG = log
        self._adapter = adap
        self._include_service_catalog = include_service_catalog
        self._requested_auth_version = requested_auth_version
        self._requested_auth_interface = requested_auth_interface

        # Built on-demand with self._request_strategy.
        self._request_strategy_obj = None

    @property
    def www_authenticate_uri(self):
        www_authenticate_uri = self._adapter.get_endpoint(
            interface=plugin.AUTH_INTERFACE)

        # NOTE(jamielennox): This weird stripping of the prefix hack is
        # only relevant to the legacy case. We urljoin '/' to get just the
        # base URI as this is the original behaviour.
        if isinstance(self._adapter.auth, _auth.AuthTokenPlugin):
            www_authenticate_uri = urllib.parse.urljoin(
                www_authenticate_uri, '/').rstrip('/')

        return www_authenticate_uri

    @property
    def auth_version(self):
        return self._request_strategy.AUTH_VERSION

    @property
    def _request_strategy(self):
        if not self._request_strategy_obj:
            strategy_class = self._get_strategy_class()
            self._adapter.version = strategy_class.AUTH_VERSION

            self._request_strategy_obj = strategy_class(
                self._adapter,
                include_service_catalog=self._include_service_catalog,
                requested_auth_interface=self._requested_auth_interface)

        return self._request_strategy_obj

    def _get_strategy_class(self):
        if self._requested_auth_version:
            if not discover.version_match(_V3RequestStrategy.AUTH_VERSION,
                                          self._requested_auth_interface):
                self._LOG.info('A version other than v3 was requested: %s',
                               self._requested_auth_interface)
            # Return v3, even if the request is unknown
            return _V3RequestStrategy

        # Specific version was not requested then we fall through to
        # discovering available versions from the server
        for klass in _REQUEST_STRATEGIES:
            if self._adapter.get_endpoint(version=klass.AUTH_VERSION):
                self._LOG.debug('Auth Token confirmed use of %s apis',
                                klass.AUTH_VERSION)
                return klass

        versions = ['v%d.%d' % s.AUTH_VERSION for s in _REQUEST_STRATEGIES]
        self._LOG.error('No attempted versions [%s] supported by server',
                        ', '.join(versions))

        msg = _('No compatible apis supported by server')
        raise ksm_exceptions.ServiceError(msg)

    def verify_token(self, user_token, retry=True, allow_expired=False):
        """Authenticate user token with identity server.

        :param user_token: user's token id
        :param retry: flag that forces the middleware to retry
                      user authentication when an indeterminate
                      response is received. Optional.
        :param allow_expired: Allow retrieving an expired token.
        :returns: access info received from identity server on success
        :rtype: :py:class:`keystoneauth1.access.AccessInfo`
        :raises exc.InvalidToken: if token is rejected
        :raises exc.ServiceError: if unable to authenticate token

        """
        try:
            auth_ref = self._request_strategy.verify_token(
                user_token,
                allow_expired=allow_expired)
        except ksa_exceptions.NotFound as e:
            self._LOG.info('Authorization failed for token')
            self._LOG.info('Identity response: %s', e.response.text)
            raise ksm_exceptions.InvalidToken(_('Token authorization failed'))
        except ksa_exceptions.Unauthorized as e:
            self._LOG.info('Identity server rejected authorization')
            self._LOG.warning('Identity response: %s', e.response.text)
            if retry:
                self._LOG.info('Retrying validation')
                return self.verify_token(user_token, False)
            msg = _('Identity server rejected authorization necessary to '
                    'fetch token data')
            raise ksm_exceptions.ServiceError(msg)
        except ksa_exceptions.HttpError as e:
            self._LOG.error(
                'Bad response code while validating token: %s %s',
                e.http_status, e.message)
            if hasattr(e.response, 'text'):
                self._LOG.warning('Identity response: %s', e.response.text)
            msg = _('Failed to fetch token data from identity server')
            raise ksm_exceptions.ServiceError(msg)
        else:
            return auth_ref

    def invalidate(self):
        return self._adapter.invalidate()



# Copyright 2022 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import http.client as http_client
import json
import logging
import ssl
from unittest import mock
import uuid
import webob.dec

import fixtures
from oslo_config import cfg
import testresources

from keystoneauth1 import access
from keystoneauth1 import exceptions as ksa_exceptions

from keystonemiddleware import oauth2_mtls_token
from keystonemiddleware.tests.unit.auth_token import base
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import BASE_URI
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import ENDPOINT_NOT_FOUND_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import ERROR_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import FAKE_ADMIN_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import FAKE_ADMIN_TOKEN_ID
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import FakeApp
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import FakeOsloCache
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import request_timeout_response
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import TIMEOUT_TOKEN
from keystonemiddleware.tests.unit.auth_token.test_auth_token_middleware \
    import VERSION_LIST_v3
from keystonemiddleware.tests.unit import client_fixtures
from keystonemiddleware.tests.unit.test_oauth2_token_middleware \
    import FakeOauth2TokenV3App
from keystonemiddleware.tests.unit.test_oauth2_token_middleware \
    import get_authorization_header
from keystonemiddleware.tests.unit import utils

_no_value = object()


class FakeSocket(object):

    def __init__(self, binary_peer_cert):
        self.binary_peer_cert = binary_peer_cert

    def getpeercert(self, binary_form=True):
        return self.binary_peer_cert


class FakeWsgiInput(object):

    def __init__(self, fake_socket):
        self.fake_socket = fake_socket

    def get_socket(self):
        return self.fake_socket


class BaseOauth2mTlsTokenMiddlewareTest(base.BaseAuthTokenTestCase):

    def setUp(self, expected_env=None, auth_version=None, fake_app=None):
        cfg.CONF.clear()
        super(BaseOauth2mTlsTokenMiddlewareTest, self).setUp()

        self.logger = self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
        self.useFixture(
            fixtures.MockPatchObject(oauth2_mtls_token.OAuth2mTlsProtocol,
                                     '_create_oslo_cache',
                                     return_value=FakeOsloCache))
        self.expected_env = expected_env or dict()
        self.fake_app = fake_app or FakeApp
        self.middleware = None

        self.conf = {
            'identity_uri': 'https://keystone.example.com:1234/testadmin/',
            'auth_version': auth_version,
            'www_authenticate_uri': 'https://keystone.example.com:1234',
            'admin_user': uuid.uuid4().hex,
        }
        self.auth_version = auth_version

    def call_middleware(self, **kwargs):
        return self.call(self.middleware, **kwargs)

    def set_middleware(self, expected_env=None, conf=None):
        """Configure the class ready to call the oauth2_token middleware.

        Set up the various fake items needed to run the middleware.
        Individual tests that need to further refine these can call this
        function to override the class defaults.

        """
        if conf:
            self.conf.update(conf)

        if expected_env:
            self.expected_env.update(expected_env)
        self.middleware = oauth2_mtls_token.OAuth2mTlsProtocol(
            self.fake_app(self.expected_env), self.conf)

    def call(self, middleware, method='GET', path='/', headers=None,
             expected_status=http_client.OK,
             expected_body_string=None, **kwargs):
        req = webob.Request.blank(path, **kwargs)
        req.method = method

        for k, v in (headers or {}).items():
            req.headers[k] = v

        resp = req.get_response(middleware)
        self.assertEqual(expected_status, resp.status_int)
        if expected_body_string:
            self.assertIn(expected_body_string, str(resp.body))
        resp.request = req
        return resp

    def assertUnauthorizedResp(self, resp):
        error = json.loads(resp.body)

        self.assertEqual('Keystone uri="https://keystone.example.com:1234"',
                         resp.headers['WWW-Authenticate'])
        self.assertEqual(
            'Keystone uri="%s"' % self.conf.get('www_authenticate_uri'),
            resp.headers['WWW-Authenticate'])
        self.assertEqual(
            'Unauthorized',
            error.get('error').get('title'))
        self.assertEqual(
            'The request you have made requires authentication.',
            error.get('error').get('message'))


class Oauth2mTlsTokenMiddlewareTest(BaseOauth2mTlsTokenMiddlewareTest,
                                    testresources.ResourcedTestCase):
    resources = [('examples', client_fixtures.EXAMPLES_RESOURCE)]

    def setUp(self):
        super(Oauth2mTlsTokenMiddlewareTest, self).setUp(
            auth_version='v3.0',
            fake_app=FakeOauth2TokenV3App)
        self.requests_mock.post('%s/v2.0/tokens' % BASE_URI,
                                text=FAKE_ADMIN_TOKEN)
        self.requests_mock.get(BASE_URI,
                               json=VERSION_LIST_v3,
                               status_code=300)
        self.requests_mock.get('%s/v3/auth/tokens' % BASE_URI,
                               text=self.token_response,
                               headers={'X-Subject-Token': uuid.uuid4().hex})
        self.set_middleware(conf={'service_type': 'tacker'})

    def token_response(self, request, context):
        auth_id = request.headers.get('X-Auth-Token')
        token_id = request.headers.get('X-Subject-Token')
        self.assertEqual(auth_id, FAKE_ADMIN_TOKEN_ID)

        if token_id == ERROR_TOKEN:
            msg = "Network connection refused."
            raise ksa_exceptions.ConnectFailure(msg)
        if token_id == ENDPOINT_NOT_FOUND_TOKEN:
            raise ksa_exceptions.EndpointNotFound()
        if token_id == TIMEOUT_TOKEN:
            request_timeout_response(request, context)

        try:
            response = self.examples.JSON_TOKEN_RESPONSES[token_id]
        except KeyError:
            response = ""
            context.status_code = 404

        return response

    def call_middleware(self, client_cert=_no_value, **kwargs):
        if client_cert is _no_value:
            client_cert = self.examples.V3_OAUTH2_MTLS_CERTIFICATE

        if client_cert:
            fake_socket = FakeSocket(client_cert)
            fake_wsgi_input = FakeWsgiInput(fake_socket)
            kwargs.update({'environ': {'wsgi.input': fake_wsgi_input}})

        return self.call(self.middleware, **kwargs)

    def test_basic(self):
        token = self.examples.v3_OAUTH2_CREDENTIAL
        token_data = self.examples.TOKEN_RESPONSES[token]

        resp = self.call_middleware(
            headers=get_authorization_header(token),
            expected_status=200,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
        )
        self.assertEqual(FakeApp.SUCCESS, resp.body)
        token_auth = resp.request.environ['keystone.token_auth']
        self.assertTrue(token_auth.has_user_token)
        self.assertEqual(token_data.user_id, token_auth.user.user_id)
        self.assertEqual(token_data.project_id, token_auth.user.project_id)
        self.assertEqual(token_data.user_domain_id,
                         token_auth.user.user_domain_id)
        self.assertEqual(token_data.project_domain_id,
                         token_auth.user.project_domain_id)
        self.assertEqual(token_data.oauth2_thumbprint,
                         token_auth.user.oauth2_credential_thumbprint)

    def test_not_oauth2_credential_token(self):
        token = self.examples.v3_APP_CRED_TOKEN
        resp = self.call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Invalid OAuth2.0 certificate-bound access token: '
            'The token is not an OAuth2.0 credential access token.',
            self.logger.output)

    def test_thumbprint_not_match(self):
        diff_cert = self.examples.V3_OAUTH2_MTLS_CERTIFICATE_DIFF
        token = self.examples.v3_OAUTH2_CREDENTIAL
        resp = self.call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
            client_cert=diff_cert
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn('The two thumbprints do not match.',
                      self.logger.output)

    @mock.patch.object(ssl, 'DER_cert_to_PEM_cert')
    def test_gen_thumbprint_exception(self, mock_DER_cert_to_PEM_cert):
        except_msg = 'Boom!'
        mock_DER_cert_to_PEM_cert.side_effect = Exception(except_msg)
        token = self.examples.v3_OAUTH2_CREDENTIAL
        resp = self.call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages'
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(except_msg, self.logger.output)

    def test_without_cert(self):
        token = self.examples.v3_OAUTH2_CREDENTIAL
        resp = self.call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
            client_cert=None
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn('Unable to obtain the client certificate.',
                      self.logger.output)

    def test_not_wsgi_input(self):
        token = self.examples.v3_OAUTH2_CREDENTIAL
        resp = super(Oauth2mTlsTokenMiddlewareTest, self).call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
            environ={'wsgi.input': None}
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn('Unable to obtain the client certificate.',
                      self.logger.output)

    def test_not_socket(self):
        token = self.examples.v3_OAUTH2_CREDENTIAL
        resp = super(Oauth2mTlsTokenMiddlewareTest, self).call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
            environ={'wsgi.input': FakeWsgiInput(None)}
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn('Unable to obtain the client certificate.',
                      self.logger.output)

    def test_not_peer_cert(self):
        token = self.examples.v3_OAUTH2_CREDENTIAL
        resp = super(Oauth2mTlsTokenMiddlewareTest, self).call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
            environ={'wsgi.input': FakeWsgiInput(FakeSocket(None))}
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn('Unable to obtain the client certificate.',
                      self.logger.output)

    @mock.patch.object(access, 'create')
    def test_keystonemiddleware_exceptiton(self, mock_create):
        except_msg = 'Unrecognized auth response'
        mock_create.side_effect = Exception(except_msg)
        token = self.examples.v3_OAUTH2_CREDENTIAL
        resp = self.call_middleware(
            headers=get_authorization_header(token),
            expected_status=401,
            method='GET', path='/vnfpkgm/v1/vnf_packages',
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Invalid token contents.',
            self.logger.output)
        self.assertIn(
            'Invalid OAuth2.0 certificate-bound access token: %s'
            % 'Token authorization failed',
            self.logger.output)

    def test_request_no_token(self):
        resp = self.call_middleware(expected_status=401)
        self.assertUnauthorizedResp(resp)

    def test_request_blank_token(self):
        resp = self.call_middleware(headers=get_authorization_header(''),
                                    expected_status=401)
        self.assertUnauthorizedResp(resp)

    def _get_cached_token(self, token):
        return self.middleware._token_cache.get(token)

    def assert_valid_last_url(self, token_id):
        self.assertLastPath('/v3/auth/tokens')

    def assertLastPath(self, path):
        if path:
            self.assertEqual(BASE_URI + path,
                             self.requests_mock.last_request.url)
        else:
            self.assertIsNone(self.requests_mock.last_request)

    def test_http_error_not_cached_token(self):
        """Test to don't cache token as invalid on network errors.

        We use UUID tokens since they are the easiest one to reach
        get_http_connection.
        """
        self.set_middleware(conf={'http_request_max_retries': '0'})
        self.call_middleware(headers=get_authorization_header(ERROR_TOKEN),
                             expected_status=503)
        self.assertIsNone(self._get_cached_token(ERROR_TOKEN))
        self.assert_valid_last_url(ERROR_TOKEN)


class FilterFactoryTest(utils.BaseTestCase):

    def test_filter_factory(self):
        conf = {}
        auth_filter = oauth2_mtls_token.filter_factory(conf)
        m = auth_filter(FakeOauth2TokenV3App())
        self.assertIsInstance(m, oauth2_mtls_token.OAuth2mTlsProtocol)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import copy
import hashlib
import os
import ssl
import time
import uuid

import jwt.utils
import oslo_cache
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
import requests.auth
import webob.dec
import webob.exc

from keystoneauth1 import exceptions as ksa_exceptions
from keystoneauth1 import loading
from keystoneauth1.loading import session as session_loading

from keystonemiddleware._common import config
from keystonemiddleware.auth_token import _cache
from keystonemiddleware.exceptions import ConfigurationError
from keystonemiddleware.exceptions import KeystoneMiddlewareException
from keystonemiddleware.i18n import _

oslo_cache.configure(cfg.CONF)
_EXT_AUTH_CONFIG_GROUP_NAME = 'ext_oauth2_auth'
_EXTERNAL_AUTH2_OPTS = [
    cfg.StrOpt('certfile',
               help='Required if identity server requires client '
                    'certificate.'),
    cfg.StrOpt('keyfile',
               help='Required if identity server requires client '
                    'private key.'),
    cfg.StrOpt('cafile',
               help='A PEM encoded Certificate Authority to use when '
                    'verifying HTTPs connections. Defaults to system CAs.'),
    cfg.BoolOpt('insecure', default=False, help='Verify HTTPS connections.'),
    cfg.IntOpt('http_connect_timeout',
               help='Request timeout value for communicating with Identity '
                    'API server.'),
    cfg.StrOpt('introspect_endpoint',
               help='The endpoint for introspect API, it is used to verify '
                    'that the OAuth 2.0 access token is valid.'),
    cfg.StrOpt('audience',
               help='The Audience should be the URL of the Authorization '
                    'Server\'s Token Endpoint. The Authorization Server will '
                    'verify that it is an intended audience for the token.'),
    cfg.StrOpt('auth_method',
               default='client_secret_basic',
               choices=('client_secret_basic', 'client_secret_post',
                        'tls_client_auth', 'private_key_jwt',
                        'client_secret_jwt'),
               help='The auth_method must use the authentication method '
                    'specified by the Authorization Server. The system '
                    'supports 5 authentication methods such as '
                    'tls_client_auth, client_secret_basic, '
                    'client_secret_post, client_secret_jwt, private_key_jwt.'),
    cfg.StrOpt('client_id',
               help='The OAuth 2.0 Client Identifier valid at the '
                    'Authorization Server.'),
    cfg.StrOpt('client_secret',
               help='The OAuth 2.0 client secret. When the auth_method is '
                    'client_secret_basic, client_secret_post, or '
                    'client_secret_jwt, the value is used, and otherwise the '
                    'value is ignored.'),
    cfg.BoolOpt('thumbprint_verify', default=False,
                help='If the access token generated by the Authorization '
                     'Server is bound to the OAuth 2.0 certificate '
                     'thumbprint, the value can be set to true, and then the '
                     'keystone middleware will verify the thumbprint.'),
    cfg.StrOpt('jwt_key_file',
               help='The jwt_key_file must use the certificate key file which '
                    'has been registered with the Authorization Server. '
                    'When the auth_method is private_key_jwt, the value is '
                    'used, and otherwise the value is ignored.'),
    cfg.StrOpt('jwt_algorithm',
               help='The jwt_algorithm must use the algorithm specified by '
                    'the Authorization Server. When the auth_method is '
                    'client_secret_jwt, this value is often set to HS256, '
                    'when the auth_method is private_key_jwt, the value is '
                    'often set to RS256, and otherwise the value is ignored.'),
    cfg.IntOpt('jwt_bearer_time_out', default=3600,
               help='This value is used to calculate the expiration time. If '
                    'after the expiration time, the access token can not be '
                    'accepted. When the auth_method is client_secret_jwt or '
                    'private_key_jwt, the value is used, and otherwise the '
                    'value is ignored.'),
    cfg.StrOpt('mapping_project_id',
               help='Specifies the method for obtaining the project ID that '
                    'currently needs to be accessed. '),
    cfg.StrOpt('mapping_project_name',
               help='Specifies the method for obtaining the project name that '
                    'currently needs to be accessed.'),
    cfg.StrOpt('mapping_project_domain_id',
               help='Specifies the method for obtaining the project domain ID '
                    'that currently needs to be accessed.'),
    cfg.StrOpt('mapping_project_domain_name',
               help='Specifies the method for obtaining the project domain '
                    'name that currently needs to be accessed.'),
    cfg.StrOpt('mapping_user_id', default='client_id',
               help='Specifies the method for obtaining the user ID.'),
    cfg.StrOpt('mapping_user_name', default='username',
               help='Specifies the method for obtaining the user name.'),
    cfg.StrOpt('mapping_user_domain_id',
               help='Specifies the method for obtaining the domain ID which '
                    'the user belongs.'),
    cfg.StrOpt('mapping_user_domain_name',
               help='Specifies the method for obtaining the domain name which '
                    'the user belongs.'),
    cfg.StrOpt('mapping_roles',
               help='Specifies the method for obtaining the list of roles in '
                    'a project or domain owned by the user.'),
    cfg.StrOpt('mapping_system_scope',
               help='Specifies the method for obtaining the scope information '
                    'indicating whether a token is system-scoped.'),
    cfg.StrOpt('mapping_expires_at',
               help='Specifies the method for obtaining the token expiration '
                    'time.'),
    cfg.ListOpt('memcached_servers',
                deprecated_name='memcache_servers',
                help='Optionally specify a list of memcached server(s) to '
                     'use for caching. If left undefined, tokens will '
                     'instead be cached in-process.'),
    cfg.IntOpt('token_cache_time',
               default=300,
               help='In order to prevent excessive effort spent validating '
                    'tokens, the middleware caches previously-seen tokens '
                    'for a configurable duration (in seconds). Set to -1 to '
                    'disable caching completely.'),
    cfg.StrOpt('memcache_security_strategy',
               default='None',
               choices=('None', 'MAC', 'ENCRYPT'),
               ignore_case=True,
               help='(Optional) If defined, indicate whether token data '
                    'should be authenticated or authenticated and encrypted. '
                    'If MAC, token data is authenticated (with HMAC) in the '
                    'cache. If ENCRYPT, token data is encrypted and '
                    'authenticated in the cache. If the value is not one of '
                    'these options or empty, auth_token will raise an '
                    'exception on initialization.'),
    cfg.StrOpt('memcache_secret_key',
               secret=True,
               help='(Optional, mandatory if memcache_security_strategy is '
                    'defined) This string is used for key derivation.'),
    cfg.IntOpt('memcache_pool_dead_retry',
               default=5 * 60,
               help='(Optional) Number of seconds memcached server is '
                    'considered dead before it is tried again.'),
    cfg.IntOpt('memcache_pool_maxsize',
               default=10,
               help='(Optional) Maximum total number of open connections to '
                    'every memcached server.'),
    cfg.IntOpt('memcache_pool_socket_timeout',
               default=3,
               help='(Optional) Socket timeout in seconds for communicating '
                    'with a memcached server.'),
    cfg.IntOpt('memcache_pool_unused_timeout',
               default=60,
               help='(Optional) Number of seconds a connection to memcached '
                    'is held unused in the pool before it is closed.'),
    cfg.IntOpt('memcache_pool_conn_get_timeout',
               default=10,
               help='(Optional) Number of seconds that an operation will wait '
                    'to get a memcached client connection from the pool.'),
    cfg.BoolOpt('memcache_use_advanced_pool',
                default=True,
                help='(Optional) Use the advanced (eventlet safe) memcached '
                     'client pool.')
]

cfg.CONF.register_opts(_EXTERNAL_AUTH2_OPTS,
                       group=_EXT_AUTH_CONFIG_GROUP_NAME)


class InvalidToken(KeystoneMiddlewareException):
    """Raise an InvalidToken Error.

    When can not get necessary information from the token,
    this error will be thrown.
    """


class ForbiddenToken(KeystoneMiddlewareException):
    """Raise a ForbiddenToken Error.

    When can not get necessary information from the token,
    this error will be thrown.
    """


class ServiceError(KeystoneMiddlewareException):
    """Raise a ServiceError.

    When can not verify any tokens, this error will be thrown.
    """


class AbstractAuthClient(object, metaclass=abc.ABCMeta):
    """Abstract http client using to access the OAuth2.0 Server."""

    def __init__(self, session, introspect_endpoint, audience, client_id,
                 func_get_config_option, logger):
        self.session = session
        self.introspect_endpoint = introspect_endpoint
        self.audience = audience
        self.client_id = client_id
        self.get_config_option = func_get_config_option
        self.logger = logger

    @abc.abstractmethod
    def introspect(self, access_token):
        """Access the introspect API."""
        pass


class ClientSecretBasicAuthClient(AbstractAuthClient):
    """Http client with the auth method 'client_secret_basic'."""

    def __init__(self, session, introspect_endpoint, audience, client_id,
                 func_get_config_option, logger):
        super(ClientSecretBasicAuthClient, self).__init__(
            session, introspect_endpoint, audience, client_id,
            func_get_config_option, logger)
        self.client_secret = self.get_config_option(
            'client_secret', is_required=True)

    def introspect(self, access_token):
        """Access the introspect API.

        Access the Introspect API to verify the access token by
        the auth method 'client_secret_basic'.
        """
        req_data = {'token': access_token,
                    'token_type_hint': 'access_token'}
        auth = requests.auth.HTTPBasicAuth(self.client_id,
                                           self.client_secret)
        http_response = self.session.request(
            self.introspect_endpoint,
            'POST',
            authenticated=False,
            data=req_data,
            requests_auth=auth)
        return http_response


class ClientSecretPostAuthClient(AbstractAuthClient):
    """Http client with the auth method 'client_secret_post'."""

    def __init__(self, session, introspect_endpoint, audience, client_id,
                 func_get_config_option, logger):
        super(ClientSecretPostAuthClient, self).__init__(
            session, introspect_endpoint, audience, client_id,
            func_get_config_option, logger)
        self.client_secret = self.get_config_option(
            'client_secret', is_required=True)

    def introspect(self, access_token):
        """Access the introspect API.

        Access the Introspect API to verify the access token by
        the auth method 'client_secret_post'.
        """
        req_data = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'token': access_token,
            'token_type_hint': 'access_token'
        }
        http_response = self.session.request(
            self.introspect_endpoint,
            'POST',
            authenticated=False,
            data=req_data)
        return http_response


class TlsClientAuthClient(AbstractAuthClient):
    """Http client with the auth method 'tls_client_auth'."""

    def introspect(self, access_token):
        """Access the introspect API.

        Access the Introspect API to verify the access token by
        the auth method 'tls_client_auth'.
        """
        req_data = {
            'client_id': self.client_id,
            'token': access_token,
            'token_type_hint': 'access_token'
        }
        http_response = self.session.request(
            self.introspect_endpoint,
            'POST',
            authenticated=False,
            data=req_data)
        return http_response


class PrivateKeyJwtAuthClient(AbstractAuthClient):
    """Http client with the auth method 'private_key_jwt'."""

    def __init__(self, session, introspect_endpoint, audience, client_id,
                 func_get_config_option, logger):
        super(PrivateKeyJwtAuthClient, self).__init__(
            session, introspect_endpoint, audience, client_id,
            func_get_config_option, logger)
        self.jwt_key_file = self.get_config_option(
            'jwt_key_file', is_required=True)
        self.jwt_bearer_time_out = self.get_config_option(
            'jwt_bearer_time_out', is_required=True)
        self.jwt_algorithm = self.get_config_option(
            'jwt_algorithm', is_required=True)
        self.logger = logger

    def introspect(self, access_token):
        """Access the introspect API.

        Access the Introspect API to verify the access token by
        the auth method 'private_key_jwt'.
        """
        if not os.path.isfile(self.jwt_key_file):
            self.logger.critical('Configuration error. JWT key file is '
                                 'not a file. path: %s' % self.jwt_key_file)
            raise ConfigurationError(_('Configuration error. '
                                       'JWT key file is not a file.'))
        try:
            with open(self.jwt_key_file, 'r') as jwt_file:
                jwt_key = jwt_file.read()
        except Exception as e:
            self.logger.critical('Configuration error. Failed to read '
                                 'the JWT key file. %s', e)
            raise ConfigurationError(_('Configuration error. '
                                       'Failed to read the JWT key file.'))
        if not jwt_key:
            self.logger.critical('Configuration error. The JWT key file '
                                 'content is empty. path: %s'
                                 % self.jwt_key_file)
            raise ConfigurationError(_('Configuration error. The JWT key file '
                                       'content is empty.'))

        iat = round(time.time())
        try:
            client_assertion = jwt.encode(
                payload={
                    'jti': str(uuid.uuid4()),
                    'iat': str(iat),
                    'exp': str(iat + self.jwt_bearer_time_out),
                    'iss': self.client_id,
                    'sub': self.client_id,
                    'aud': self.audience},
                headers={
                    'typ': 'JWT',
                    'alg': self.jwt_algorithm},
                key=jwt_key,
                algorithm=self.jwt_algorithm)
        except Exception as e:
            self.logger.critical('Configuration error. JWT encoding with '
                                 'the specified JWT key file and algorithm '
                                 'failed. path: %s, algorithm: %s, error: %s' %
                                 (self.jwt_key_file, self.jwt_algorithm, e))
            raise ConfigurationError(_('Configuration error. JWT encoding '
                                       'with the specified JWT key file '
                                       'and algorithm failed.'))
        req_data = {
            'client_id': self.client_id,
            'client_assertion_type':
                'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': client_assertion,
            'token': access_token,
            'token_type_hint': 'access_token'
        }
        http_response = self.session.request(
            self.introspect_endpoint,
            'POST',
            authenticated=False,
            data=req_data)
        return http_response


class ClientSecretJwtAuthClient(AbstractAuthClient):
    """Http client with the auth method 'client_secret_jwt'."""

    def __init__(self, session, introspect_endpoint, audience, client_id,
                 func_get_config_option, logger):
        super(ClientSecretJwtAuthClient, self).__init__(
            session, introspect_endpoint, audience, client_id,
            func_get_config_option, logger)
        self.client_secret = self.get_config_option(
            'client_secret', is_required=True)
        self.jwt_bearer_time_out = self.get_config_option(
            'jwt_bearer_time_out', is_required=True)
        self.jwt_algorithm = self.get_config_option(
            'jwt_algorithm', is_required=True)

    def introspect(self, access_token):
        """Access the introspect API.

        Access the Introspect API to verify the access token by
        the auth method 'client_secret_jwt'.
        """
        ita = round(time.time())
        try:
            client_assertion = jwt.encode(
                payload={
                    'jti': str(uuid.uuid4()),
                    'iat': str(ita),
                    'exp': str(ita + self.jwt_bearer_time_out),
                    'iss': self.client_id,
                    'sub': self.client_id,
                    'aud': self.audience},
                headers={
                    'typ': 'JWT',
                    'alg': self.jwt_algorithm},
                key=self.client_secret,
                algorithm=self.jwt_algorithm)
        except Exception as e:
            self.logger.critical('Configuration error. JWT encoding with '
                                 'the specified client_secret and algorithm '
                                 'failed. algorithm: %s, error: %s'
                                 % (self.jwt_algorithm, e))
            raise ConfigurationError(_('Configuration error. JWT encoding '
                                       'with the specified client_secret '
                                       'and algorithm failed.'))
        req_data = {
            'client_id': self.client_id,
            'client_assertion_type':
                'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': client_assertion,
            'token': access_token,
            'token_type_hint': 'access_token'
        }
        http_response = self.session.request(
            self.introspect_endpoint,
            'POST',
            authenticated=False,
            data=req_data)
        return http_response


_ALL_AUTH_CLIENTS = {
    'client_secret_basic': ClientSecretBasicAuthClient,
    'client_secret_post': ClientSecretPostAuthClient,
    'tls_client_auth': TlsClientAuthClient,
    'private_key_jwt': PrivateKeyJwtAuthClient,
    'client_secret_jwt': ClientSecretJwtAuthClient
}


def _get_http_client(auth_method, session, introspect_endpoint, audience,
                     client_id, func_get_config_option, logger):
    """Get an auth HTTP Client to access the OAuth2.0 Server."""
    if auth_method in _ALL_AUTH_CLIENTS:
        return _ALL_AUTH_CLIENTS.get(auth_method)(
            session, introspect_endpoint, audience,
            client_id, func_get_config_option, logger)
    logger.critical('The value is incorrect for option '
                    'auth_method in group [%s]' %
                    _EXT_AUTH_CONFIG_GROUP_NAME)
    raise ConfigurationError(_('The configuration parameter for '
                               'key "auth_method" in group [%s] '
                               'is incorrect.') %
                             _EXT_AUTH_CONFIG_GROUP_NAME)


class ExternalAuth2Protocol(object):
    """Middleware that handles External Server OAuth2.0 authentication."""

    def __init__(self, application, conf):
        super(ExternalAuth2Protocol, self).__init__()
        self._application = application
        self._log = logging.getLogger(conf.get('log_name', __name__))
        self._log.info('Starting Keystone external_oauth2_token middleware')

        config_opts = [
            (_EXT_AUTH_CONFIG_GROUP_NAME, _EXTERNAL_AUTH2_OPTS
             + loading.get_auth_common_conf_options())
        ]
        all_opts = [(g, copy.deepcopy(o)) for g, o in config_opts]
        self._conf = config.Config('external_oauth2_token',
                                   _EXT_AUTH_CONFIG_GROUP_NAME,
                                   all_opts,
                                   conf)
        self._token_cache = self._token_cache_factory()

        self._session = self._create_session()
        self._audience = self._get_config_option('audience', is_required=True)
        self._introspect_endpoint = self._get_config_option(
            'introspect_endpoint', is_required=True)
        self._auth_method = self._get_config_option(
            'auth_method', is_required=True)
        self._client_id = self._get_config_option(
            'client_id', is_required=True)
        self._http_client = _get_http_client(
            self._auth_method, self._session, self._introspect_endpoint,
            self._audience, self._client_id,
            self._get_config_option, self._log)

    def _token_cache_factory(self):
        security_strategy = self._conf.get('memcache_security_strategy')
        cache_kwargs = dict(
            cache_time=int(self._conf.get('token_cache_time')),
            memcached_servers=self._conf.get('memcached_servers'),
            use_advanced_pool=self._conf.get(
                'memcache_use_advanced_pool'),
            dead_retry=self._conf.get('memcache_pool_dead_retry'),
            maxsize=self._conf.get('memcache_pool_maxsize'),
            unused_timeout=self._conf.get(
                'memcache_pool_unused_timeout'),
            conn_get_timeout=self._conf.get(
                'memcache_pool_conn_get_timeout'),
            socket_timeout=self._conf.get(
                'memcache_pool_socket_timeout'),
        )
        if security_strategy.lower() != 'none':
            secret_key = self._conf.get('memcache_secret_key')
            return _cache.SecureTokenCache(self._log,
                                           security_strategy,
                                           secret_key,
                                           **cache_kwargs)
        return _cache.TokenCache(self._log, **cache_kwargs)

    @webob.dec.wsgify()
    def __call__(self, req):
        """Handle incoming request."""
        self.process_request(req)
        response = req.get_response(self._application)
        return self.process_response(response)

    def process_request(self, request):
        """Process request.

        :param request: Incoming request
        :type request: _request.AuthTokenRequest
        """
        access_token = None
        if (request.authorization and
                request.authorization.authtype == 'Bearer'):
            access_token = request.authorization.params

        try:
            if not access_token:
                self._log.info('Unable to obtain the access token.')
                raise InvalidToken(_('Unable to obtain the access token.'))

            self._token_cache.initialize(request.environ)
            token_data = self._fetch_token(access_token)

            if (self._get_config_option('thumbprint_verify',
                                        is_required=False)):
                self._confirm_certificate_thumbprint(
                    request, token_data.get('origin_token_metadata'))

            self._set_request_env(request, token_data)

        except InvalidToken as error:
            self._log.info('Rejecting request. '
                           'Need a valid OAuth 2.0 access token. '
                           'error: %s', error)
            message = _('The request you have made is denied, '
                        'because the token is invalid.')
            body = {'error': {
                'code': 401,
                'title': 'Unauthorized',
                'message': message,
            }}
            raise webob.exc.HTTPUnauthorized(
                body=jsonutils.dumps(body),
                headers=self._reject_headers,
                charset='UTF-8',
                content_type='application/json')
        except ForbiddenToken as error:
            self._log.warning('Rejecting request. '
                              'The necessary information is required.'
                              'error: %s', error)
            message = _('The request you have made is denied, '
                        'because the necessary information '
                        'could not be parsed.')
            body = {'error': {
                'code': 403,
                'title': 'Forbidden',
                'message': message,
            }}
            raise webob.exc.HTTPForbidden(
                body=jsonutils.dumps(body),
                charset='UTF-8',
                content_type='application/json')
        except ConfigurationError as error:
            self._log.critical('Rejecting request. '
                               'The configuration parameters are incorrect. '
                               'error: %s', error)
            message = _('The request you have made is denied, '
                        'because the configuration parameters are incorrect '
                        'and the token can not be verified.')
            body = {'error': {
                'code': 500,
                'title': 'Internal Server Error',
                'message': message,
            }}
            raise webob.exc.HTTPServerError(
                body=jsonutils.dumps(body),
                charset='UTF-8',
                content_type='application/json')
        except ServiceError as error:
            self._log.warning('Rejecting request. An exception occurred and '
                              'the OAuth 2.0 access token can not be '
                              'verified. error: %s', error)
            message = _('The request you have made is denied, '
                        'because an exception occurred while accessing '
                        'the external authentication server '
                        'for token validation.')
            body = {'error': {
                'code': 500,
                'title': 'Internal Server Error',
                'message': message,
            }}
            raise webob.exc.HTTPServerError(
                body=jsonutils.dumps(body),
                charset='UTF-8',
                content_type='application/json')

    def process_response(self, response):
        """Process Response.

        Add ``WWW-Authenticate`` headers to requests that failed with
        ``401 Unauthenticated`` so users know where to authenticate for future
        requests.
        """
        if response.status_int == 401:
            response.headers.extend(self._reject_headers)
        return response

    def _create_session(self, **kwargs):
        """Create session for HTTP access."""
        kwargs.setdefault('cert', self._get_config_option(
            'certfile', is_required=False))
        kwargs.setdefault('key', self._get_config_option(
            'keyfile', is_required=False))
        kwargs.setdefault('cacert', self._get_config_option(
            'cafile', is_required=False))
        kwargs.setdefault('insecure', self._get_config_option(
            'insecure', is_required=False))
        kwargs.setdefault('timeout', self._get_config_option(
            'http_connect_timeout', is_required=False))
        kwargs.setdefault('user_agent', self._conf.user_agent)
        return session_loading.Session().load_from_options(**kwargs)

    def _get_config_option(self, key, is_required):
        """Read the value from config file by the config key."""
        value = self._conf.get(key)
        if not value:
            if is_required:
                self._log.critical('The value is required for option %s '
                                   'in group [%s]' % (
                                       key, _EXT_AUTH_CONFIG_GROUP_NAME))
                raise ConfigurationError(
                    _('Configuration error. The parameter '
                      'is not set for "%s" in group [%s].') % (
                        key, _EXT_AUTH_CONFIG_GROUP_NAME))
            else:
                return None
        else:
            return value

    @property
    def _reject_headers(self):
        """Generate WWW-Authenticate Header.

        When response status is 401, this method will be called to add
        the 'WWW-Authenticate' header to the response.
        """
        header_val = 'Authorization OAuth 2.0 uri="%s"' % self._audience
        return [('WWW-Authenticate', header_val)]

    def _fetch_token(self, access_token):
        """Use access_token to get the valid token meta_data.

        Verify the access token through accessing the external
        authorization server.
        """
        try:
            cached = self._token_cache.get(access_token)
            if cached:
                self._log.debug('The cached token: %s' % cached)
                if (not isinstance(cached, dict)
                        or 'origin_token_metadata' not in cached):
                    self._log.warning('The cached data is invalid. %s' %
                                      cached)
                    raise InvalidToken(_('The token is invalid.'))
                origin_token_metadata = cached.get('origin_token_metadata')
                if not origin_token_metadata.get('active'):
                    self._log.warning('The cached data is invalid. %s' %
                                      cached)
                    raise InvalidToken(_('The token is invalid.'))
                expire_at = self._read_data_from_token(
                    origin_token_metadata, 'mapping_expires_at',
                    is_required=False, value_type=int)
                if expire_at:
                    if int(expire_at) < int(time.time()):
                        cached['origin_token_metadata']['active'] = False
                        self._token_cache.set(access_token, cached)
                        self._log.warning(
                            'The cached data is invalid. %s' % cached)
                        raise InvalidToken(_('The token is invalid.'))
                return cached
            http_response = self._http_client.introspect(access_token)
            if http_response.status_code != 200:
                self._log.critical('The introspect API returns an '
                                   'incorrect response. '
                                   'response_status: %s, response_text: %s' %
                                   (http_response.status_code,
                                    http_response.text))
                raise ServiceError(_('The token cannot be verified '
                                     'for validity.'))

            origin_token_metadata = http_response.json()
            self._log.debug('The introspect API response: %s' %
                            origin_token_metadata)
            if not origin_token_metadata.get('active'):
                self._token_cache.set(
                    access_token,
                    {'origin_token_metadata': origin_token_metadata})
                self._log.info('The token is invalid. response: %s' %
                               origin_token_metadata)
                raise InvalidToken(_('The token is invalid.'))
            token_data = self._parse_necessary_info(origin_token_metadata)
            self._token_cache.set(access_token, token_data)
            return token_data

        except (ConfigurationError, ForbiddenToken,
                ServiceError, InvalidToken):
            raise
        except (ksa_exceptions.ConnectFailure,
                ksa_exceptions.DiscoveryFailure,
                ksa_exceptions.RequestTimeout) as error:
            self._log.critical('Unable to validate token: %s', error)
            raise ServiceError(
                _('The Introspect API service is temporarily unavailable.'))
        except Exception as error:
            self._log.critical('Unable to validate token: %s', error)
            raise ServiceError(_('An exception occurred during the token '
                                 'verification process.'))

    def _read_data_from_token(self, token_metadata, config_key,
                              is_required=False, value_type=None):
        """Read value from token metadata.

        Read the necessary information from the token metadata with the
        config key.
        """
        if not value_type:
            value_type = str
        meta_key = self._get_config_option(config_key, is_required=is_required)
        if not meta_key:
            return None

        if meta_key.find('.') >= 0:
            meta_value = None
            for temp_key in meta_key.split('.'):
                if not temp_key:
                    self._log.critical('Configuration error. '
                                       'config_key: %s , meta_key: %s ' %
                                       (config_key, meta_key))
                    raise ConfigurationError(
                        _('Failed to parse the necessary information '
                          'for the field "%s".') % meta_key)
                if not meta_value:
                    meta_value = token_metadata.get(temp_key)
                else:
                    if not isinstance(meta_value, dict):
                        self._log.warning(
                            'Failed to parse the necessary information. '
                            'The meta_value is not of type dict.'
                            'config_key: %s , meta_key: %s, value: %s' %
                            (config_key, meta_key, meta_value))
                        raise ForbiddenToken(
                            _('Failed to parse the necessary information '
                              'for the field "%s".') % meta_key)
                    meta_value = meta_value.get(temp_key)
        else:
            meta_value = token_metadata.get(meta_key)

        if not meta_value:
            if is_required:
                self._log.warning(
                    'Failed to parse the necessary information. '
                    'The meta value is required.'
                    'config_key: %s , meta_key: %s, value: %s, need_type: %s' %
                    (config_key, meta_key, meta_value, value_type))
                raise ForbiddenToken(_('Failed to parse the necessary '
                                       'information for the field "%s".') %
                                     meta_key)
            else:
                meta_value = None
        else:
            if not isinstance(meta_value, value_type):
                self._log.warning(
                    'Failed to parse the necessary information. '
                    'The meta value is of an incorrect type.'
                    'config_key: %s , meta_key: %s, value: %s, need_type: %s'
                    % (config_key, meta_key, meta_value, value_type))
                raise ForbiddenToken(_('Failed to parse the necessary '
                                       'information for the field "%s".') %
                                     meta_key)
        return meta_value

    def _parse_necessary_info(self, token_metadata):
        """Parse the necessary information from the token metadata."""
        token_data = dict()
        token_data['origin_token_metadata'] = token_metadata

        roles = self._read_data_from_token(token_metadata,
                                           'mapping_roles',
                                           is_required=True)
        is_admin = 'false'
        if 'admin' in roles.lower().split(','):
            is_admin = 'true'
        token_data['roles'] = roles
        token_data['is_admin'] = is_admin

        system_scope = self._read_data_from_token(
            token_metadata, 'mapping_system_scope',
            is_required=False, value_type=bool)
        if system_scope:
            token_data['system_scope'] = 'all'
        else:
            project_id = self._read_data_from_token(
                token_metadata, 'mapping_project_id', is_required=False)
            if project_id:
                token_data['project_id'] = project_id
                token_data['project_name'] = self._read_data_from_token(
                    token_metadata, 'mapping_project_name', is_required=True)
                token_data['project_domain_id'] = self._read_data_from_token(
                    token_metadata, 'mapping_project_domain_id',
                    is_required=True)
                token_data['project_domain_name'] = self._read_data_from_token(
                    token_metadata, 'mapping_project_domain_name',
                    is_required=True)
            else:
                token_data['domain_id'] = self._read_data_from_token(
                    token_metadata, 'mapping_project_domain_id',
                    is_required=True)
                token_data['domain_name'] = self._read_data_from_token(
                    token_metadata, 'mapping_project_domain_name',
                    is_required=True)

        token_data['user_id'] = self._read_data_from_token(
            token_metadata, 'mapping_user_id', is_required=True)
        token_data['user_name'] = self._read_data_from_token(
            token_metadata, 'mapping_user_name', is_required=True)
        token_data['user_domain_id'] = self._read_data_from_token(
            token_metadata, 'mapping_user_domain_id', is_required=True)
        token_data['user_domain_name'] = self._read_data_from_token(
            token_metadata, 'mapping_user_domain_name', is_required=True)

        return token_data

    def _get_client_certificate(self, request):
        """Get the client certificate from request environ or socket."""
        try:
            pem_client_cert = request.environ.get('SSL_CLIENT_CERT')
            if pem_client_cert:
                peer_cert = ssl.PEM_cert_to_DER_cert(pem_client_cert)
            else:
                wsgi_input = request.environ.get('wsgi.input')
                if not wsgi_input:
                    self._log.warn('Unable to obtain the client certificate. '
                                   'The object for wsgi_input is none.')
                    raise InvalidToken(_('Unable to obtain the client '
                                         'certificate.'))
                socket = wsgi_input.get_socket()
                if not socket:
                    self._log.warn('Unable to obtain the client certificate. '
                                   'The object for socket is none.')
                    raise InvalidToken(_('Unable to obtain the client '
                                         'certificate.'))
                peer_cert = socket.getpeercert(binary_form=True)
            if not peer_cert:
                self._log.warn('Unable to obtain the client certificate. '
                               'The object for peer_cert is none.')
                raise InvalidToken(_('Unable to obtain the client '
                                     'certificate.'))
            return peer_cert
        except InvalidToken:
            raise
        except Exception as error:
            self._log.warn('Unable to obtain the client certificate. %s' %
                           error)
            raise InvalidToken(_('Unable to obtain the client certificate.'))

    def _confirm_certificate_thumbprint(self, request, origin_token_metadata):
        """Check if the thumbprint in the token is valid."""
        peer_cert = self._get_client_certificate(request)
        try:
            thumb_sha256 = hashlib.sha256(peer_cert).digest()
            cert_thumb = jwt.utils.base64url_encode(thumb_sha256).decode(
                'ascii')
        except Exception as error:
            self._log.warn('An Exception occurred. %s' % error)
            raise InvalidToken(_('Can not generate the thumbprint.'))

        token_thumb = origin_token_metadata.get('cnf', {}).get('x5t#S256')
        if cert_thumb != token_thumb:
            self._log.warn('The two thumbprints do not match. '
                           'token_thumbprint: %s, certificate_thumbprint %s' %
                           (token_thumb, cert_thumb))
            raise InvalidToken(_('The two thumbprints do not match.'))

    def _set_request_env(self, request, token_data):
        """Set request.environ with the necessary information."""
        request.environ['external.token_info'] = token_data
        request.environ['HTTP_X_IDENTITY_STATUS'] = 'Confirmed'
        request.environ['HTTP_X_ROLES'] = token_data.get('roles')
        request.environ['HTTP_X_ROLE'] = token_data.get('roles')
        request.environ['HTTP_X_USER_ID'] = token_data.get('user_id')
        request.environ['HTTP_X_USER_NAME'] = token_data.get('user_name')
        request.environ['HTTP_X_USER_DOMAIN_ID'] = token_data.get(
            'user_domain_id')
        request.environ['HTTP_X_USER_DOMAIN_NAME'] = token_data.get(
            'user_domain_name')
        if token_data.get('is_admin') == 'true':
            request.environ['HTTP_X_IS_ADMIN_PROJECT'] = token_data.get(
                'is_admin')
        request.environ['HTTP_X_USER'] = token_data.get('user_name')

        if token_data.get('system_scope'):
            request.environ['HTTP_OPENSTACK_SYSTEM_SCOPE'] = token_data.get(
                'system_scope'
            )
        elif token_data.get('project_id'):
            request.environ['HTTP_X_PROJECT_ID'] = token_data.get('project_id')
            request.environ['HTTP_X_PROJECT_NAME'] = token_data.get(
                'project_name')
            request.environ['HTTP_X_PROJECT_DOMAIN_ID'] = token_data.get(
                'project_domain_id')
            request.environ['HTTP_X_PROJECT_DOMAIN_NAME'] = token_data.get(
                'project_domain_name')
            request.environ['HTTP_X_TENANT_ID'] = token_data.get('project_id')
            request.environ['HTTP_X_TENANT_NAME'] = token_data.get(
                'project_name')
            request.environ['HTTP_X_TENANT'] = token_data.get('project_id')
        else:
            request.environ['HTTP_X_DOMAIN_ID'] = token_data.get('domain_id')
            request.environ['HTTP_X_DOMAIN_NAME'] = token_data.get(
                'domain_name')
        self._log.debug('The access token data is %s.' % jsonutils.dumps(
            token_data))


def filter_factory(global_conf, **local_conf):
    """Return a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_filter(app):
        return ExternalAuth2Protocol(app, conf)

    return auth_filter



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from keystoneauth1 import loading
from oslo_config import cfg

from keystonemiddleware.auth_token import _base


# NOTE(jamielennox): A number of options below are deprecated however are left
# in the list and only mentioned as deprecated in the help string. This is
# because we have to provide the same deprecation functionality for arguments
# passed in via the conf in __init__ (from paste) and there is no way to test
# that the default value was set or not in CONF.
# Also if we were to remove the options from the CONF list (as typical CONF
# deprecation works) then other projects will not be able to override the
# options via CONF.

_OPTS = [
    cfg.StrOpt('www_authenticate_uri',
               # FIXME(dolph): should be default='http://127.0.0.1:5000/v2.0/',
               # or (depending on client support) an unversioned, publicly
               # accessible identity endpoint (see bug 1207517). Further, we
               # can eliminate this configuration option in favor of pulling
               # the endpoint from the service catalog that the service user
               # receives (there should be an identity endpoint listed there).
               # This wasn't an option originally when many auth_token
               # deployments were configured with the "ADMIN" token and
               # endpoint combination.
               deprecated_name='auth_uri',
               help='Complete "public" Identity API endpoint. This endpoint'
               ' should not be an "admin" endpoint, as it should be accessible'
               ' by all end users. Unauthenticated clients are redirected to'
               ' this endpoint to authenticate. Although this endpoint should'
               ' ideally be unversioned, client support in the wild varies.'
               ' If you\'re using a versioned v2 endpoint here, then this'
               ' should *not* be the same endpoint the service user utilizes'
               ' for validating tokens, because normal end users may not be'
               ' able to reach that endpoint.'),
    cfg.StrOpt('auth_uri',
               deprecated_for_removal=True,
               deprecated_reason='The auth_uri option is deprecated in favor'
               ' of www_authenticate_uri and will be removed in the S '
               ' release.',
               deprecated_since='Queens',
               help='Complete "public" Identity API endpoint. This endpoint'
               ' should not be an "admin" endpoint, as it should be accessible'
               ' by all end users. Unauthenticated clients are redirected to'
               ' this endpoint to authenticate. Although this endpoint should'
               ' ideally be unversioned, client support in the wild varies.'
               ' If you\'re using a versioned v2 endpoint here, then this'
               ' should *not* be the same endpoint the service user utilizes'
               ' for validating tokens, because normal end users may not be'
               ' able to reach that endpoint. This option is deprecated in'
               ' favor of www_authenticate_uri and will be removed in the S'
               ' release.'),
    cfg.StrOpt('auth_version',
               help='API version of the Identity API endpoint.'),
    cfg.StrOpt('interface',
               default='internal',
               help='Interface to use for the Identity API endpoint. Valid'
               ' values are "public", "internal" (default) or "admin".'),
    cfg.BoolOpt('delay_auth_decision',
                default=False,
                help='Do not handle authorization requests within the'
                ' middleware, but delegate the authorization decision to'
                ' downstream WSGI components.'),
    cfg.IntOpt('http_connect_timeout',
               help='Request timeout value for communicating with Identity'
               ' API server.'),
    cfg.IntOpt('http_request_max_retries',
               default=3,
               help='How many times are we trying to reconnect when'
               ' communicating with Identity API Server.'),
    cfg.StrOpt('cache',
               help='Request environment key where the Swift cache object is'
               ' stored. When auth_token middleware is deployed with a Swift'
               ' cache, use this option to have the middleware share a caching'
               ' backend with swift. Otherwise, use the ``memcached_servers``'
               ' option instead.'),
    cfg.StrOpt('certfile',
               help='Required if identity server requires client certificate'),
    cfg.StrOpt('keyfile',
               help='Required if identity server requires client certificate'),
    cfg.StrOpt('cafile',
               help='A PEM encoded Certificate Authority to use when '
                    'verifying HTTPs connections. Defaults to system CAs.'),
    cfg.BoolOpt('insecure', default=False, help='Verify HTTPS connections.'),
    cfg.StrOpt('region_name',
               help='The region in which the identity server can be found.'),
    cfg.ListOpt('memcached_servers',
                deprecated_name='memcache_servers',
                help='Optionally specify a list of memcached server(s) to'
                ' use for caching. If left undefined, tokens will instead be'
                ' cached in-process.'),
    cfg.IntOpt('token_cache_time',
               default=300,
               help='In order to prevent excessive effort spent validating'
               ' tokens, the middleware caches previously-seen tokens for a'
               ' configurable duration (in seconds). Set to -1 to disable'
               ' caching completely.'),
    cfg.StrOpt('memcache_security_strategy',
               default='None',
               choices=('None', 'MAC', 'ENCRYPT'),
               ignore_case=True,
               help='(Optional) If defined, indicate whether token data'
               ' should be authenticated or authenticated and encrypted.'
               ' If MAC, token data is authenticated (with HMAC) in the cache.'
               ' If ENCRYPT, token data is encrypted and authenticated in the'
               ' cache. If the value is not one of these options or empty,'
               ' auth_token will raise an exception on initialization.'),
    cfg.StrOpt('memcache_secret_key',
               secret=True,
               help='(Optional, mandatory if memcache_security_strategy is'
               ' defined) This string is used for key derivation.'),
    cfg.IntOpt('memcache_pool_dead_retry',
               default=5 * 60,
               help='(Optional) Number of seconds memcached server is'
               ' considered dead before it is tried again.'),
    cfg.IntOpt('memcache_pool_maxsize',
               default=10,
               help='(Optional) Maximum total number of open connections to'
               ' every memcached server.'),
    cfg.IntOpt('memcache_pool_socket_timeout',
               default=3,
               help='(Optional) Socket timeout in seconds for communicating '
                    'with a memcached server.'),
    cfg.IntOpt('memcache_pool_unused_timeout',
               default=60,
               help='(Optional) Number of seconds a connection to memcached'
               ' is held unused in the pool before it is closed.'),
    cfg.IntOpt('memcache_pool_conn_get_timeout',
               default=10,
               help='(Optional) Number of seconds that an operation will wait '
                    'to get a memcached client connection from the pool.'),
    cfg.BoolOpt('memcache_use_advanced_pool',
                default=True,
                help='(Optional) Use the advanced (eventlet safe) memcached '
                     'client pool.'),
    cfg.BoolOpt('include_service_catalog',
                default=True,
                help='(Optional) Indicate whether to set the X-Service-Catalog'
                ' header. If False, middleware will not ask for service'
                ' catalog on token validation and will not set the'
                ' X-Service-Catalog header.'),
    cfg.StrOpt('enforce_token_bind',
               default='permissive',
               help='Used to control the use and type of token binding. Can'
               ' be set to: "disabled" to not check token binding.'
               ' "permissive" (default) to validate binding information if the'
               ' bind type is of a form known to the server and ignore it if'
               ' not. "strict" like "permissive" but if the bind type is'
               ' unknown the token will be rejected. "required" any form of'
               ' token binding is needed to be allowed. Finally the name of a'
               ' binding method that must be present in tokens.'),
    cfg.ListOpt('service_token_roles', default=['service'],
                help='A choice of roles that must be present in a service'
                ' token. Service tokens are allowed to request that an expired'
                ' token can be used and so this check should tightly control'
                ' that only actual services should be sending this token.'
                ' Roles here are applied as an ANY check so any role in this'
                ' list must be present. For backwards compatibility reasons'
                ' this currently only affects the allow_expired check.'),
    cfg.BoolOpt('service_token_roles_required', default=False,
                help='For backwards compatibility reasons we must let valid'
                ' service tokens pass that don\'t pass the service_token_roles'
                ' check as valid. Setting this true will become the default'
                ' in a future release and should be enabled if possible.'),
    cfg.StrOpt('service_type',
               help='The name or type of the service as it appears in the'
               ' service catalog. This is used to validate tokens that have'
               ' restricted access rules.'),
    cfg.BoolOpt('memcache_sasl_enabled',
                default=False,
                help='Enable the SASL(Simple Authentication and Security'
                     ' Layer) if the SASL_enable is true, else disable.'),
    cfg.StrOpt('memcache_username',
               default='',
               help='the user name for the SASL'),
    cfg.StrOpt('memcache_password',
               default='',
               help='the username password for SASL'),
]


CONF = cfg.CONF
CONF.register_opts(_OPTS, group=_base.AUTHTOKEN_GROUP)
loading.register_auth_conf_options(cfg.CONF, _base.AUTHTOKEN_GROUP)


auth_token_opts = [
    (_base.AUTHTOKEN_GROUP, _OPTS + loading.get_auth_common_conf_options()),
]

__all__ = (
    'list_opts',
)


def list_opts():
    """Return a list of oslo_config options available in auth_token middleware.

    The returned list includes the non-deprecated oslo_config options which may
    be registered at runtime by the project. The purpose of this is to allow
    tools like the Oslo sample config file generator to discover the options
    exposed to users by this middleware.

    Deprecated Options should not show up here so as to not be included in
    sample configuration.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    This function is discoverable via the entry point
    'keystonemiddleware.auth_token' under the 'oslo.config.opts' namespace.

    :returns: a list of (group_name, opts) tuples
    """
    auth_token_opts = (_OPTS + loading.get_auth_common_conf_options())

    return [(_base.AUTHTOKEN_GROUP, copy.deepcopy(auth_token_opts))]



# Copyright 2019 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystonemiddleware.auth_token import _path_matches
from keystonemiddleware.tests.unit import utils


class TestAccessRules(utils.BaseTestCase):

    def test_path_matches(self):
        good_matches = [
            ('/v2/servers', '/v2/servers'),
            ('/v2/servers/123', '/v2/servers/{server_id}'),
            ('/v2/servers/123/', '/v2/servers/{server_id}/'),
            ('/v2/servers/123', '/v2/servers/*'),
            ('/v2/servers/123/', '/v2/servers/*/'),
            ('/v2/servers/123', '/v2/servers/**'),
            ('/v2/servers/123/', '/v2/servers/**'),
            ('/v2/servers/123/456', '/v2/servers/**'),
            ('/v2/servers', '**'),
            ('/v2/servers/', '**'),
            ('/v2/servers/123', '**'),
            ('/v2/servers/123/456', '**'),
            ('/v2/servers/123/volume/456', '**'),
            ('/v2/servers/123/456', '/v2/*/*/*'),
            ('/v2/123/servers/466', '/v2/{project_id}/servers/{server_id}'),
        ]
        for (request, pattern) in good_matches:
            self.assertIsNotNone(_path_matches(request, pattern))
        bad_matches = [
            ('/v2/servers/someuuid', '/v2/servers'),
            ('/v2/servers//', '/v2/servers/{server_id}'),
            ('/v2/servers/123/', '/v2/servers/{server_id}'),
            ('/v2/servers/123/456', '/v2/servers/{server_id}'),
            ('/v2/servers/123/456', '/v2/servers/*'),
            ('/v2/servers', 'v2/servers'),
            ('/v2/servers/123/456/789', '/v2/*/*/*'),
            ('/v2/servers/123/', '/v2/*/*/*'),
            ('/v2/servers/', '/v2/servers/{server_id}'),
            ('/v2/servers', '/v2/servers/{server_id}'),
        ]
        for (request, pattern) in bad_matches:
            self.assertIsNone(_path_matches(request, pattern))
